from typing import Any

import pyspark.sql.functions as F
from pyspark.sql import DataFrame


def filter_on_range(
    df: DataFrame,
    col: str,
    from_: Any,
    to: Any,
    include_from: bool = True,
    include_to: bool = True,
) -> DataFrame:
    if include_from:
        from_cdt = F.col(col) >= from_
    else:
        from_cdt = F.col(col) > from_
    if include_to:
        to_cdt = F.col(col) <= to
    else:
        to_cdt = F.col(col) < to

    return df.filter((from_cdt & to_cdt))



from typing import Optional

import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from pyspark.sql.types import DateType, TimestampType


def with_date_column(
    df: DataFrame, col: str, fmt: str, new_col: Optional[str] = None
) -> DataFrame:
    if new_col is None:
        new_col = col

    return df.withColumn(
        new_col,
        F.unix_timestamp(F.col(col), format=fmt).cast(TimestampType()).cast(DateType()),
    )



from typing import Any, Collection, List, Union

import pyspark.sql.functions as F
from pyspark.sql import DataFrame


def has_unique_id(df: DataFrame, cols: Union[str, List[str]]) -> bool:
    return df.count() == df.select(cols).distinct().count()


def has_column(df: DataFrame, col: str) -> bool:
    return col in df.columns


def has_columns(df: DataFrame, cols: Union[str, List[str]]) -> bool:
    if isinstance(cols, str):
        cols = [cols]
    return all((col in df.columns for col in cols))


def has_distinct_values_in_set(
    df: DataFrame, cols: Union[str, List[str]], values: Collection[Any]
) -> bool:
    try:
        df.expect_distinct_values_in_set(cols=cols, values=values)  # type: ignore[operator]
    except Exception:
        return False
    else:
        return True


def has_distinct_values_equal_set(
    df: DataFrame, cols: Union[str, List[str]], values: Collection[Any]
) -> bool:
    try:
        df.expect_distinct_values_equal_set(cols=cols, values=values)  # type: ignore[operator]
    except Exception:
        return False
    else:
        return True


def has_min_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> bool:
    return _check_value_between(df=df, cols=cols, min=min, max=max, fn=F.min)


def has_avg_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> bool:
    return _check_value_between(df=df, cols=cols, min=min, max=max, fn=F.avg)


def has_max_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> bool:
    return _check_value_between(df=df, cols=cols, min=min, max=max, fn=F.max)


def _check_value_between(
    df: DataFrame,
    cols: Union[str, List[str]],
    min: Any,
    max: Any,
    fn,
):
    if isinstance(cols, str):
        cols = [cols]
    for col in cols:
        found_value = df.select(fn(col).alias(col)).collect()[0][col]
        if not (min <= found_value <= max):
            return False
    return True



from pyspark.sql import SparkSession


def get_spark_session() -> SparkSession:
    """Get or create a SparkSession.

    >>> assert(isinstance(get_spark_session(), SparkSession))
    """
    return SparkSession.builder.getOrCreate()



from pyspark.sql import DataFrame


def schema_equals(df1: DataFrame, df2: DataFrame, check_nullable: bool):
    def field_list(fields):
        return fields.name, fields.dataType, fields.nullable

    fields1 = [*map(field_list, df1.schema.fields)]
    fields2 = [*map(field_list, df2.schema.fields)]
    if check_nullable:
        res = set(fields1) == set(fields2)
    else:
        res = {field[:-1] for field in fields1} == {field[:-1] for field in fields2}
    return res


def data_equals(df1: DataFrame, df2: DataFrame, sort_cols: bool):
    if sort_cols:
        df1 = df1.select(sorted(df1.columns))
        df2 = df2.select(sorted(df2.columns))

    data1 = df1.collect()
    data2 = df2.collect()
    return set(data1) == set(data2)


def df_equals(df1: DataFrame, df2: DataFrame, check_nullable: bool, sort_cols: bool):
    schema = schema_equals(df1, df2, check_nullable=check_nullable)
    data = data_equals(df1, df2, sort_cols=sort_cols)
    return schema and data



import pyspark.sql.functions as F
from pyspark.sql import DataFrame


def drop_col_if_na(df: DataFrame, max: int = 0) -> DataFrame:
    """Returns a new DataFrame with columns dropped if NA values were found.

    Args:
        df: DataFrame
        max: threshold of allowed NA values within one column, default: 0

    Returns: new DataFrame with columns dropped
    """
    col_counts = [F.count(F.when(F.col(c).isNull(), c)).alias(c) for c in df.columns]
    counts = df.select(col_counts).collect()[0].asDict()
    to_drop = [column for column, cnt in counts.items() if cnt > max]
    return df.drop(*to_drop)


def drop_col_if_not_distinct(df: DataFrame, min: int = 2) -> DataFrame:
    """Returns a new DataFrame with columns dropped if values are not distinct.

    Args:
        df: DataFrame
        min: threshold of distinct values within one column, default: 2

    Returns: new DataFrame with columns dropped
    """
    col_counts = [F.countDistinct(c).alias(c) for c in df.columns]
    counts = df.select(col_counts).collect()[0].asDict()
    to_drop = [column for column, cnt in counts.items() if cnt < min]
    return df.drop(*to_drop)



from pyspark.sql import DataFrame

from silex.fn.date import with_date_column
from silex.fn.drop import drop_col_if_na, drop_col_if_not_distinct
from silex.fn.expect import (
    expect_avg_value_between,
    expect_column,
    expect_columns,
    expect_distinct_values_equal_set,
    expect_distinct_values_in_set,
    expect_max_value_between,
    expect_min_value_between,
    expect_unique_id,
)
from silex.fn.filter import filter_on_range
from silex.fn.has import (
    has_avg_value_between,
    has_column,
    has_columns,
    has_distinct_values_equal_set,
    has_distinct_values_in_set,
    has_max_value_between,
    has_min_value_between,
    has_unique_id,
)
from silex.fn.join import join_closest_date


def extend_dataframes():
    fns = [
        drop_col_if_na,
        drop_col_if_not_distinct,
        expect_avg_value_between,
        expect_column,
        expect_columns,
        expect_distinct_values_equal_set,
        expect_distinct_values_in_set,
        expect_max_value_between,
        expect_min_value_between,
        expect_unique_id,
        filter_on_range,
        has_avg_value_between,
        has_column,
        has_columns,
        has_distinct_values_equal_set,
        has_distinct_values_in_set,
        has_max_value_between,
        has_min_value_between,
        has_unique_id,
        join_closest_date,
        with_date_column,
    ]

    for fn in fns:
        setattr(DataFrame, fn.__name__, fn)  # noqa: B010



from typing import List

import pyspark.sql.functions as F
from pyspark.sql import DataFrame, Window


def join_closest_date(
    df: DataFrame,
    other: DataFrame,
    id_cols: List[str],
    date_col: str,
    other_date_col: str,
    date_diff_col: str = "date_diff",
    dense_rank_col: str = "dr",
    keep_date_diff: bool = False,
) -> DataFrame:
    window = Window.partitionBy(id_cols).orderBy(F.abs(date_diff_col))

    res = (
        df.join(other, id_cols)
        .withColumn(date_diff_col, F.datediff(date_col, other_date_col))
        .withColumn(dense_rank_col, F.dense_rank().over(window))
        .where(f"{dense_rank_col}=1")  # select closest date
        .drop(dense_rank_col)
    )
    if not keep_date_diff:
        return res.drop(date_diff_col)
    return res



from contextlib import suppress
from typing import Union


def parse(s: str) -> Union[None, str, bool, int, float]:
    """Try to cast to:
        1. None: <null> (case is ignored)
        2. bool: {<true>, <false>} (case is ignored)
        3. integer
        4. float

    or keep as string.

    Returns: None

    >>> parse("<null>") is None
    True
    >>> parse("<true>")
    True
    >>> parse("<FALSE>")
    False
    >>> parse("1")
    1
    >>> parse("1.0")
    1.0
    >>> parse("a")
    'a'
    """
    s_upper = s.upper()
    if s_upper == "<NULL>":
        return None
    if s_upper == "<TRUE>":
        return True
    if s_upper == "<FALSE>":
        return False
    with suppress(ValueError):
        return int(s)
    with suppress(ValueError):
        return float(s)
    return s



from typing import Any, Collection, List, Union

import pyspark.sql.functions as F
from pyspark.sql import DataFrame

from silex.exception import (
    SilexDuplicateIdsException,
    SilexMissingColumnException,
    SilexMissingColumnsException,
    SilexMissingValuesException,
    SilexUnexpectedValueException,
    SilexUnexpectedValuesException,
)


def expect_unique_id(df: DataFrame, cols: Union[str, List[str]]) -> DataFrame:
    if not df.has_unique_id(cols):  # type: ignore[operator]
        raise SilexDuplicateIdsException(cols)
    return df


def expect_column(df: DataFrame, col: str) -> DataFrame:
    if not df.has_column(col):  # type: ignore[operator]
        raise SilexMissingColumnException(col)
    return df


def expect_columns(df: DataFrame, cols: Union[str, List[str]]) -> DataFrame:
    missing_cols = set()
    if isinstance(cols, str):
        cols = [cols]
    for col in cols:
        if col not in df.columns:
            missing_cols.add(col)
    if len(missing_cols) > 0:
        raise SilexMissingColumnsException(missing_cols)
    return df


def expect_distinct_values_in_set(
    df: DataFrame,
    cols: Union[str, List[str]],
    values: Collection[Any],
    err_msg: str = "value outside of set",
) -> DataFrame:
    if isinstance(cols, str):
        cols = [cols]
    assertions = [
        F.assert_true(F.col(col).isin(list(set(values))), errMsg=err_msg)
        for col in cols
    ]
    try:
        df.select(*assertions).collect()
    except Exception as e:
        raise SilexUnexpectedValuesException(cols=cols) from e
    else:
        return df


def expect_distinct_values_equal_set(
    df: DataFrame, cols: Union[str, List[str]], values: Collection[Any]
) -> DataFrame:
    if isinstance(cols, str):
        cols = [cols]
    distinct_v = set()
    for col in cols:
        distinct_v.update([row[col] for row in df.select(col).distinct().collect()])
        if not distinct_v.issubset(values):
            raise SilexUnexpectedValuesException(col)

    if distinct_v != set(values):
        raise SilexMissingValuesException(found=distinct_v, expected=values)
    return df


def expect_min_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> DataFrame:
    if not df.has_min_value_between(cols=cols, min=min, max=max):  # type: ignore[operator]
        raise SilexUnexpectedValueException()
    return df


def expect_avg_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> DataFrame:
    if not df.has_avg_value_between(cols=cols, min=min, max=max):  # type: ignore[operator]
        raise SilexUnexpectedValueException()
    return df


def expect_max_value_between(
    df: DataFrame, cols: Union[str, List[str]], min: Any, max: Any
) -> DataFrame:
    if not df.has_max_value_between(cols=cols, min=min, max=max):  # type: ignore[operator]
        raise SilexUnexpectedValueException()
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['silex', 'silex.fn', 'silex.testing', 'silex.utils']

package_data = \
{'': ['*']}

install_requires = \
['pyspark>=3']

setup_kwargs = {
    'name': 'spark-silex',
    'version': '0.2.0',
    'description': 'Silex adds more sparks to your project!',
    'long_description': '# Silex\n\nAdd more 🔥 to [Apache Spark](https://spark.apache.org/)!\n\n[![Python](https://img.shields.io/badge/Python3.8-Python?style=for-the-badge&logo=Python)](https://www.python.org/downloads/release/python-380/)\n[![Manager: Poetry](https://img.shields.io/badge/Manager-Poetry-blue?style=for-the-badge)](https://python-poetry.org/)\n[![Test: BDD](https://img.shields.io/badge/Test-BDD-critical?style=for-the-badge)](https://github.com/behave/behave)\n[![Test: Doctest](https://img.shields.io/badge/Test-Doctest-success?style=for-the-badge)](https://docs.python.org/3/library/doctest.html)\n\n[![Code style: Black](https://img.shields.io/badge/Codestyle-Black-black?style=for-the-badge)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=for-the-badge&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Linter: Flake8](https://img.shields.io/badge/Linter-Flake8-black?style=for-the-badge)](https://github.com/PyCQA/flake8)\n[![try/except style: tryceratops](https://img.shields.io/badge/try%2Fexcept%20style-tryceratops%20%F0%9F%A6%96%E2%9C%A8-black?style=for-the-badge)](https://github.com/guilatrova/tryceratops)\n\n[![Typing: MyPy](https://img.shields.io/badge/Typing-MyPy-blue?style=for-the-badge)](https://github.com/python/mypy)\n[![Security: Bandit](https://img.shields.io/badge/Security-Bandit-critical?style=for-the-badge)](https://github.com/PyCQA/bandit)\n\n\n[![Git: Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&style=for-the-badge&logoColor=white)](https://pre-commit.com/)\n[![Git: Conventional](https://img.shields.io/badge/Git-conventional-ff69b4?style=for-the-badge)](https://www.conventionalcommits.org)\n[![Version: Semantic](https://img.shields.io/badge/Version-Semantic-black?style=for-the-badge)](https://semver.org/)\n\n![Apache Spark](https://img.shields.io/static/v1?style=for-the-badge&message=Apache+Spark+%20+3&color=E25A1C&logo=Apache+Spark&logoColor=FFFFFF&label=)\n\n## TLDR\n\nSilex is a Data Engineering library to extend PySpark.\n\nYou don\'t need another class, just use PySpark as usual and you have new functions to your DataFrames!\n\n```python\nimport silex\nfrom pyspark.sql import DataFrame\n\n# extends your DataFrame with silex functions!\n# if for some reason you don\'t want to do that, check \'Without extending Dataframes\' README section below\nsilex.extend_dataframes()\n\ndf: DataFrame = ...  # your regular Spark DataFrame\ndf: DataFrame = df.drop_col_if_na()  # new function! and still a regular Spark Dataframe!\n# scroll for more information!\n```\n\n### Available functions\n\n```python\n# assertions (raises an Exception if not met /!\\)\ndef expect_column(self, col: str) -> DataFrame: ...\ndef expect_columns(self, cols: Union[str, List[str]]) -> DataFrame: ...\n\ndef expect_distinct_values_equal_set(self, cols: Union[str, List[str]], values: Collection[Any]) -> DataFrame: ...\ndef expect_distinct_values_in_set(self, cols: Union[str, List[str]], values: Collection[Any]) -> DataFrame: ...\n\ndef expect_min_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> DataFrame: ...\ndef expect_avg_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> DataFrame: ...\ndef expect_max_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> DataFrame: ...\n\ndef expect_unique_id(self, cols: Union[str, List[str]]) -> DataFrame: ...\n\n# boolean checks\ndef has_column(self, col: str) -> bool: ...\ndef has_columns(self, cols: Union[str, List[str]]) -> bool: ...\n\ndef has_distinct_values_equal_set(self, cols: Union[str, List[str]], values: Collection[Any]) -> bool: ...\ndef has_distinct_values_in_set(self, cols: Union[str, List[str]], values: Collection[Any]) -> bool: ...\n\ndef has_min_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> bool: ...\ndef has_avg_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> bool: ...\ndef has_max_value_between(self, cols: Union[str, List[str]], min: Any, max: Any) -> bool: ...\n\ndef has_unique_id(self, cols: Union[str, List[str]]) -> bool: ...\n\n# dates\ndef with_date_column(self, col: str, fmt: str, new_col: Optional[str] = None) -> DataFrame: ...\n\n# drop\ndef drop_col_if_na(self, max: int) -> DataFrame: ...\ndef drop_col_if_not_distinct(self, min: int) -> DataFrame: ...\n\n# filters\ndef filter_on_range(self, col: str, from_: Any, to: Any, ...) -> DataFrame: ...\n\n# joins\ndef join_closest_date(self, other: DataFrame, ...) -> DataFrame: ...\n```\n\n## Getting started\n\n### Pre-requisites\n\n- Python 3.8 or above\n- Spark 3 or above\n\n### Installation\n\n```shell\npip install < # TODO >\n```\n\n### Usage\n\n#### By extending DataFrames! ⚡\n\n```python\nimport silex\nfrom pyspark.sql import DataFrame, SparkSession\n\n# extends your DataFrame with silex functions!\n# if for some reason you don\'t want to do that, check next example\nsilex.extend_dataframes()\n\nspark = SparkSession.builder.getOrCreate()\n\ndata = [\n    (0, "2022-01-01", "a", 1.0),\n    (1, "2022-02-01", "b", 2.0),\n    (2, "2022-03-01", "c", 3.0),\n]\ndf: DataFrame = spark.createDataFrame(data, schema=["id", "date", "text", "value"])\n\ndf.show()\n# +---+----------+----+-----+\n# | id|      date|text|value|\n# +---+----------+----+-----+\n# |  0|2022-01-01|   a|  1.0|\n# |  1|2022-02-01|   b|  2.0|\n# |  2|2022-03-01|   c|  3.0|\n# +---+----------+----+-----+\n\ndf.printSchema()\n# root\n#  |-- id: long (nullable = true)\n#  |-- date: string (nullable = true)\n#  |-- text: string (nullable = true)\n#  |-- value: double (nullable = true)\n\ndf = df.with_date_column(col="date", fmt="yyyy-MM-dd")\n\ndf.show()\n# +---+----------+----+-----+\n# | id|      date|text|value|\n# +---+----------+----+-----+\n# |  0|2022-01-01|   a|  1.0|\n# |  1|2022-02-01|   b|  2.0|\n# |  2|2022-03-01|   c|  3.0|\n# +---+----------+----+-----+\n\ndf.printSchema()\n# root\n#  |-- id: long (nullable = true)\n#  |-- date: date (nullable = true)\n#  |-- text: string (nullable = true)\n#  |-- value: double (nullable = true)\n```\n\n#### Without extending Dataframes 🌧️\n\n```python\nfrom silex.fn.date import with_date_column\nfrom pyspark.sql import DataFrame, SparkSession\n\nspark = SparkSession.builder.getOrCreate()\n\ndata = [\n    (0, "2022-01-01", "a", 1.0),\n    (1, "2022-02-01", "b", 2.0),\n    (2, "2022-03-01", "c", 3.0),\n]\ndf: DataFrame = spark.createDataFrame(data, schema=["id", "date", "text", "value"])\n\ndf.show()\n# +---+----------+----+-----+\n# | id|      date|text|value|\n# +---+----------+----+-----+\n# |  0|2022-01-01|   a|  1.0|\n# |  1|2022-02-01|   b|  2.0|\n# |  2|2022-03-01|   c|  3.0|\n# +---+----------+----+-----+\n\ndf.printSchema()\n# root\n#  |-- id: long (nullable = true)\n#  |-- date: string (nullable = true)\n#  |-- text: string (nullable = true)\n#  |-- value: double (nullable = true)\n\ndf = with_date_column(df=df, col="date", fmt="yyyy-MM-dd")\n\ndf.show()\n# +---+----------+----+-----+\n# | id|      date|text|value|\n# +---+----------+----+-----+\n# |  0|2022-01-01|   a|  1.0|\n# |  1|2022-02-01|   b|  2.0|\n# |  2|2022-03-01|   c|  3.0|\n# +---+----------+----+-----+\n\ndf.printSchema()\n# root\n#  |-- id: long (nullable = true)\n#  |-- date: date (nullable = true)\n#  |-- text: string (nullable = true)\n#  |-- value: double (nullable = true)\n```\n\n## Contributing\n\n```shell\n# install poetry and python 3.8, using pyenv for instance\n\ncd silex\npoetry env use path/to/python3.8  # e.g. ~/.pyenv/versions/3.8.12/bin/python\npoetry shell\npoetry install\npre-commit install\n\nmake help\n# or open Makefile to learn about available commands for development\n```\n',
    'author': 'RickCi',
    'author_email': 'rci-git@pm.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/RickCi/silex',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from typing import List, Optional


class SilexException(Exception):
    pass


class SilexExpectException(SilexException):
    pass


class SilexDuplicateIdsException(SilexExpectException):
    def __init__(self, cols):
        super().__init__(f"Found duplicate ids for columns: {cols}")


class SilexMissingColumnException(SilexExpectException):
    def __init__(self, col):
        super().__init__(f"Missing column: {col}")


class SilexMissingColumnsException(SilexExpectException):
    def __init__(self, cols):
        super().__init__(f"Missing columns: {cols}")


class SilexUnexpectedValuesException(SilexExpectException):
    def __init__(self, col: Optional[str] = None, cols: Optional[List[str]] = None):
        if col:
            super().__init__(f"Unexpected values for column: {col}")
        elif cols:
            super().__init__(f"Unexpected values for columns: {cols}")
        else:
            super().__init__("Unexpected values")


class SilexUnexpectedValueException(SilexExpectException):
    def __init__(self):
        super().__init__("Unexpected value")


class SilexMissingValuesException(SilexExpectException):
    def __init__(self, found, expected):
        msg = f"set of values are not equal: len(found)={len(found)} != len(expected)={len(expected)}"
        super().__init__(msg)
